#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/udp.h>
#include <linux/netfilter_ipv4/ip_conntrack_protocol.h>
#include "./gemtek_added/UDP_Session_Idle_Time.h"

#ifdef	CONFIG_NETFILTER_DEBUG
#define DEBUGP(format, args...) printk(KERN_DEBUG format, ## args)
#else
#define DEBUGP(format, args...)
#endif

#define ISAKMP_TIMEOUT (5*HZ)
#define StopCCN_TIMEOUT (5*HZ)

#define UDP_TIMEOUT (30*HZ)
#define UDP_STREAM_TIMEOUT (180*HZ)

static unsigned short	fakeTID=1;	//This is for L2TP

static int udp_pkt_to_tuple(const void *datah, size_t datalen,
			    struct ip_conntrack_tuple *tuple)
{
	const struct udphdr *hdr = datah;
	unsigned char *l2tph = datah + sizeof (struct udphdr), *p;
	unsigned short	*PktLength, left; // pkt length not processed yet

	unsigned short *TID;
	//u_int16_t	*cookie = datah+sizeof(struct udphdr);

	switch ( ntohs(hdr->dest) ) {		// Depending on different protocol on UDP
		//case	500: 	// This is ISAKMP
			//tuple->src.u.udp.port = *cookie ^ *(cookie+1) ^ *(cookie+2) ^ *(cookie+3);
			//tuple->dst.u.udp.port = 500;
			//break;
		case	1701: 	// This is L2TP
			tuple->src.u.udp.port = hdr->source;
	// checking whether the length field exists, ChungLiang 01/20/2004 	
			if ( *(l2tph) & 0x40 )
			TID = (unsigned short *) (l2tph + 4) ;
			else 
			   TID = (unsigned short *) (l2tph + 2) ;
			if ( *TID == 0 ) { // this is SCCRQ
				 PktLength = (unsigned short *) (l2tph+2);
				 left = ntohs(*PktLength);
				 DEBUGP(KERN_DEBUG "SCCRQ found with length=%d.\n",left);
				 left -=12;
				 p = l2tph+13; //current pointer

				 // Searching for TID record
				   while ( left > 0 ) {
						 DEBUGP(KERN_DEBUG "Process AVP : %d.\n",*(p+4));
						 if ( *(p+4) == 9 ) { // this is tunnel ID AVP
							   tuple->src.u.udp.realTID = *(p+6) + (*(p+5)*256);
							    DEBUGP(KERN_DEBUG "real=%d,", *(p+6) + (*(p+5)*256) );
			/* marked off by ChungLiang for disabling Tunnel ID remap */	
			
							   tuple->src.u.udp.fakeTID = fakeTID;
							    DEBUGP(KERN_DEBUG "fake=%d.\n", fakeTID);
							    *(p+5) = fakeTID >> 8;
							    *(p+6) = fakeTID & 0xFF;
								if (hdr->check) {
				    if ((unsigned int)p & 0x0001) // this is odd align
								hdr->check = ip_nat_cheat_check(
									htons(~tuple->src.u.udp.realTID), 
									htons(fakeTID), hdr->check);
								else
								hdr->check = ip_nat_cheat_check(
									~tuple->src.u.udp.realTID, 
									fakeTID, hdr->check);
				    }
							    fakeTID++;
			
                                  break;
						 }
						 left -= *p;
						 p += *p;
				   }
			}
			else {	//We got destination TID from packet
				  tuple->src.u.udp.realTID = 0;
				  tuple->src.u.udp.fakeTID = 0;
			}
			tuple->dst.u.udp.port = hdr->dest;
			tuple->dst.u.udp.fakeTID = ntohs(*TID);	//This will be zero at the SCCRQ packet
			tuple->dst.u.udp.realTID = ntohs(*TID);

			//Following is for finding the connetion disconnect
			/*
			if (*l2tph & 0x80) {
				 PktLength = (unsigned short *) (l2tph+2);
				 left = ntohs(*PktLength);
				 left -=12;
				 p = l2tph+13; //current pointer

				 // Searching for TID record
				   while ( left > 0 ) {
						 DEBUGP(KERN_DEBUG "Process AVP : %d.\n",*(p+4));
						 if ( *(p+4) == 4 ) { // this is StopCCN 
							    DEBUGP(KERN_DEBUG "StopCCN\n");
								ip_ct_refresh(conntrack, StopCCN_TIMEOUT);
                                  break;
						 }
						 left -= *p;
						 p += *p;
				   }
			}
			*/
			break;
		default:
			tuple->src.u.udp.port = hdr->source;
			tuple->dst.u.udp.port = hdr->dest;
	}

	return 1;
}

static int udp_invert_tuple(struct ip_conntrack_tuple *tuple,
			    const struct ip_conntrack_tuple *orig)
{
	switch ( ntohs(orig->dst.u.udp.port) ) {		// Depending on different protocol on UDP
		//case	500: 	// This is ISAKMP
			//tuple->src.u.udp.port = orig->src.u.udp.port;
			//tuple->dst.u.udp.port = orig->dst.u.udp.port;
			//break;
		case	1701: 	// This is L2TP
			tuple->src.u.udp.port = orig->dst.u.udp.port;
			tuple->src.u.udp.fakeTID = orig->dst.u.udp.fakeTID;
			tuple->src.u.udp.realTID = orig->dst.u.udp.realTID;

			tuple->dst.u.udp.port = orig->src.u.udp.port;
			tuple->dst.u.udp.fakeTID = orig->src.u.udp.fakeTID;
			tuple->dst.u.udp.realTID = orig->src.u.udp.realTID;
			break;
		default:
			tuple->src.u.udp.port = orig->dst.u.udp.port;
			tuple->dst.u.udp.port = orig->src.u.udp.port;
	}
	return 1;
}

/* Print out the per-protocol part of the tuple. */
static unsigned int udp_print_tuple(char *buffer,
				    const struct ip_conntrack_tuple *tuple)
{
	switch ( ntohs(tuple->dst.u.udp.port) ) {		// Depending on different protocol on UDP
		case	1701: 	// This is L2TP
			return sprintf(buffer, "srealTID=%hu ,drealTID=%hu ",
					  tuple->src.u.udp.realTID,
					  tuple->dst.u.udp.realTID);
			break;
		default:
			return sprintf(buffer, "sport=%hu dport=%hu ",
					  ntohs(tuple->src.u.udp.port),
					  ntohs(tuple->dst.u.udp.port));
	}
}

/* Print out the private part of the conntrack. */
static unsigned int udp_print_conntrack(char *buffer,
					const struct ip_conntrack *conntrack)
{
	return 0;
}

/* Returns verdict for packet, and may modify conntracktype */
static int udp_packet(struct ip_conntrack *conntrack,
		      struct iphdr *iph, size_t len,
		      enum ip_conntrack_info conntrackinfo)
{
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)iph + iph->ihl);

	DEBUGP("UDP Port number=%u.\n",ntohs(hdr->dest) );

	if (ntohs(hdr->dest) == 500) { // This is ISAKMP
		if (conntrack->status & IPS_SEEN_REPLY) 
			set_bit(IPS_ASSURED_BIT, &conntrack->status);

		// ip_ct_refresh(conntrack, ISAKMP_TIMEOUT);
		UDP_Session_Idle_Time_Extra_SRCs_001
		return NF_ACCEPT;
	}

	/* If we've seen traffic both ways, this is some kind of UDP
	   stream.  Extend timeout. */
	if (conntrack->status & IPS_SEEN_REPLY) {
		// ip_ct_refresh(conntrack, UDP_STREAM_TIMEOUT);
		UDP_Session_Idle_Time_Extra_SRCs_001
		/* Also, more likely to be important, and not a probe */
		set_bit(IPS_ASSURED_BIT, &conntrack->status);
	} else {
		// ip_ct_refresh(conntrack, UDP_TIMEOUT);
		UDP_Session_Idle_Time_Extra_SRCs_001
	}

	return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static int udp_new(struct ip_conntrack *conntrack,
			     struct iphdr *iph, size_t len)
{
	return 1;
}

struct ip_conntrack_protocol ip_conntrack_protocol_udp
= { { NULL, NULL }, IPPROTO_UDP, "udp",
    udp_pkt_to_tuple, udp_invert_tuple, udp_print_tuple, udp_print_conntrack,
    udp_packet, udp_new, NULL };
