/*
 * Copyright 2001-2005, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <dirent.h>

#include <bcmnvram.h>
#include <shutils.h>
#include <rc.h>
#include <netconf.h>

extern int netconf_clamp_mss_to_pmtu(void);

/* Add filter to both INPUT and FORWARD tables */
static void
add_filter(netconf_filter_t *filter)
{
	int inbound[] = { NETCONF_IN, NETCONF_FORWARD, -1 }, i;
	
	for (i = 0; inbound[i] != -1; i++) {
		filter->dir = inbound[i];
		netconf_add_filter(filter);
	}
}

/* Combination PREROUTING DNAT and <dir> ACCEPT */
static void
parse_port_forward(int ipproto, int dir, int target, char *wordlist)
{
	netconf_nat_t nat;
	netconf_filter_t filter;
	char var[256], *next;
	char *wan_port0, *wan_port1, *lan_ipaddr, *lan_port0, *lan_port1, *proto, *enable;

	/* Parse wan_port0-wan_port1>lan_ipaddr:lan_port0-lan_port1:proto:enable ... */
	foreach(var, wordlist, next) {
		lan_ipaddr = var;
		wan_port0 = strsep(&lan_ipaddr, ">");
		if (!wan_port0 || !lan_ipaddr)
			continue;
		wan_port1 = wan_port0;
		wan_port0 = strsep(&wan_port1, "-");
		if (!wan_port0 || !wan_port1)
			continue;
		lan_port0 = lan_ipaddr;
		lan_ipaddr = strsep(&lan_port0, ":");
		if (!lan_ipaddr || !lan_port0)
			continue;
		lan_port1 = lan_port0;
		lan_port0 = strsep(&lan_port1, "-");
		if (!lan_port0 || !lan_port1)
			continue;
		proto = lan_port1;
		lan_port1 = strsep(&proto, ":");
		if (proto) {
			if (!strncmp(proto, "tcp", 3))
				ipproto = IPPROTO_TCP;
			else if (!strncmp(proto, "udp", 3))
				ipproto = IPPROTO_UDP;
		}
		enable = proto;
		proto = strsep(&enable, ":");
		if (enable && !strcmp(enable, "off"))
			continue;

		/* Set up WAN side match */
		memset(&nat, 0, sizeof(nat));
		nat.match.ipproto = ipproto;
		nat.match.src.ports[1] = htons(0xffff);
		nat.match.dst.ports[0] = htons(atoi(wan_port0));
		nat.match.dst.ports[1] = htons(atoi(wan_port1));
		strncpy(nat.match.in.name, nvram_safe_get("wan_ifname"), IFNAMSIZ);

		/* Set up DNAT to LAN */
		nat.target = NETCONF_DNAT;
		inet_aton(lan_ipaddr, &nat.ipaddr);
		nat.ports[0] = htons(atoi(lan_port0));
		nat.ports[1] = htons(atoi(lan_port1));

		/* Set up LAN side match */
		memset(&filter, 0, sizeof(filter));
		filter.match.ipproto = ipproto;
		filter.match.src.ports[1] = htons(0xffff);
		filter.match.dst.ipaddr.s_addr = nat.ipaddr.s_addr;
		filter.match.dst.netmask.s_addr = htonl(0xffffffff);
		filter.match.dst.ports[0] = nat.ports[0];
		filter.match.dst.ports[1] = nat.ports[1];
		strncpy(filter.match.in.name, nvram_safe_get("wan_ifname"), IFNAMSIZ);

		/* Accept connection */
		filter.target = target;
		filter.dir = dir;

		/* Do it */
		netconf_add_nat(&nat);
		netconf_add_filter(&filter);
	}
}

/* <dir> DROP */
void
parse_port_filter(int ipproto, int target, char *wordlist)
{
	netconf_filter_t filter;
	char var[256], *next;
	char *lan_ipaddr0, *lan_ipaddr1, *lan_port0, *lan_port1;
	char *proto, *enable, *day_start, *day_end, *sec_start, *sec_end;
	struct in_addr start_addr, end_addr, addr;

	/* Set up LAN side match */
	memset(&filter, 0, sizeof(filter));
	filter.target = target;
	filter.match.src.ports[1] = htons(0xffff);
	strncpy(filter.match.in.name, nvram_safe_get("lan_ifname"), IFNAMSIZ);

	/* Parse [lan_ipaddr0-lan_ipaddr1|*]:lan_port0-lan_port1:proto:enable:day_start-day_end,sec_start-sec_end ... */
	foreach(var, wordlist, next) {
		lan_port0 = var;
		lan_ipaddr0 = strsep(&lan_port0, ":");
		if (!lan_ipaddr0 || !lan_port0)
			continue;
		lan_ipaddr1 = lan_ipaddr0; 
		lan_ipaddr0 = strsep(&lan_ipaddr1, "-");
		if (!lan_ipaddr0)
			continue;
		lan_port1 = lan_port0;
		lan_port0 = strsep(&lan_port1, "-");
		if (!lan_port0 || !lan_port1)
			continue;
		proto = lan_port1;
		lan_port1 = strsep(&proto, ":");
		if (proto) {
			if (!strncmp(proto, "tcp", 3))
				ipproto = IPPROTO_TCP;
			else if (!strncmp(proto, "udp", 3))
				ipproto = IPPROTO_UDP;
		}
		enable = proto;
		proto = strsep(&enable, ":");
		if (enable && !strcmp(enable, "off"))
			continue;
		day_start = enable;
		enable = strsep(&day_start, ":");
		day_end = day_start;
		day_start = strsep(&day_end, "-");
		sec_start = day_end;
		day_end = strsep(&sec_start, ",");
		sec_end = sec_start;
		sec_start = strsep(&sec_end, "-");

		if (day_start && day_end && sec_start && sec_end) {
		}

		/* Finish LAN side match */
		filter.match.ipproto = ipproto;
		filter.match.dst.ports[0] = htons(atoi(lan_port0));
		filter.match.dst.ports[1] = htons(atoi(lan_port1));
		if (*lan_ipaddr0 == '*') {
			/* Match any IP address */
			filter.match.src.ipaddr.s_addr = htonl(0);
			filter.match.src.netmask.s_addr = htonl(0);
			add_filter(&filter);
		} else if (lan_ipaddr1) {
			/* Match a range of IP addresses */
			inet_aton(lan_ipaddr0, &start_addr);
			inet_aton(lan_ipaddr1, &end_addr);

			for (addr.s_addr = start_addr.s_addr; 
			     ntohl(addr.s_addr) <= ntohl(end_addr.s_addr);
			     addr.s_addr = htonl(ntohl(addr.s_addr) + 1)) {
				filter.match.src.ipaddr.s_addr = addr.s_addr;
				filter.match.src.netmask.s_addr = htonl(0xffffffff);
				add_filter(&filter);
			}
		} else {
			/* Match a single IP address */
			inet_aton(lan_ipaddr0, &filter.match.src.ipaddr);
			filter.match.src.netmask.s_addr = htonl(0xffffffff);
			add_filter(&filter);
		}
	}
}

int	start_firewall ( void )
{
	char *WorkingModel="0", *c;
	DIR *dir;
	struct dirent *file;
	FILE *fp;
	char name[NAME_MAX], *value;
	netconf_nat_t nat;
	netconf_filter_t filter;
	int i;
	char var[256], *next;
	int log_level, log_drop, log_accept;

	/* Reset firewall */
//	netconf_reset_fw ();

	/* Block obviously spoofed IP addresses */
	if (!(dir = opendir("/proc/sys/net/ipv4/conf")))
		perror("/proc/sys/net/ipv4/conf");
	while (dir && (file = readdir(dir))) {
		if (strncmp(file->d_name, ".", NAME_MAX) != 0 &&
		    strncmp(file->d_name, "..", NAME_MAX) != 0) {
			sprintf(name, "/proc/sys/net/ipv4/conf/%s/rp_filter", file->d_name);
			if (!(fp = fopen(name, "r+"))) {
				perror(name);
				break;
			}
			fputc('1', fp);
			fclose(fp);
		}
	}

	/* Optionally log connections */
	log_level = atoi(nvram_safe_get("log_level"));
	log_drop = (log_level & 1) ? NETCONF_LOG_DROP : NETCONF_DROP;
	log_accept = (log_level & 2) ? NETCONF_LOG_ACCEPT : NETCONF_ACCEPT;

	/*
	 * Inbound drops
	 */

	/* Drop invalid packets */
	memset(&filter, 0, sizeof(filter));
	filter.match.state = NETCONF_INVALID;
	filter.target = NETCONF_DROP;
	add_filter(&filter);

	/*
	 * Forward drops
	 */

	/* Filter by IP address */
	memset(&filter, 0, sizeof(filter));
	filter.target = log_drop;
	filter.dir = NETCONF_FORWARD;
	filter.match.src.netmask.s_addr = htonl(0xffffffff);

	/* Parse lan_ipaddr0-lan_ipaddr1 ... */
	foreach(var, nvram_safe_get("filter_ip"), next) {
		char *lan_ipaddr0, *lan_ipaddr1;
		struct in_addr start_addr, end_addr, addr;

		lan_ipaddr1 = var;
		lan_ipaddr0 = strsep(&lan_ipaddr1, "-");
		if (!lan_ipaddr0 || !lan_ipaddr1)
			continue;

		inet_aton(lan_ipaddr0, &start_addr);
		inet_aton(lan_ipaddr1, &end_addr);

		for (addr.s_addr = start_addr.s_addr; 
		     ntohl(addr.s_addr) <= ntohl(end_addr.s_addr);
		     addr.s_addr = htonl(ntohl(addr.s_addr) + 1)) {
			filter.match.src.ipaddr.s_addr = addr.s_addr;
			add_filter(&filter);
		}
	}

	/* Filter by destination ports */
	parse_port_filter(IPPROTO_TCP, NETCONF_FORWARD, nvram_safe_get("filter_tcp"));
	parse_port_filter(IPPROTO_UDP, NETCONF_FORWARD, nvram_safe_get("filter_udp"));
/*
	if ( nvram_safe_get ( "DMZEnableFlag" ) != NULL ) {
		if ( nvram_match ( "DMZEnableFlag" , "1" ) ) {
			if ( nvram_invmatch ( "dmz_ipaddr" , "" ) ) {
				snprintf ( name , sizeof ( name ) , "0-65535>%s:0-65535" , nvram_safe_get ( "dmz_ipaddr" ) );
				parse_port_forward ( IPPROTO_TCP , NETCONF_FORWARD , log_accept , name );
				parse_port_forward ( IPPROTO_UDP , NETCONF_FORWARD , log_accept , name );
			}
		} else if ( nvram_match ( "DMZEnableFlag" , "0" ) ) { ; }
		else { ; }
	} else if ( nvram_safe_get ( "DMZEnableFlag" ) == NULL ) { ; }
	else { ; }
*/

	/* Filter by MAC address */
	if (!nvram_match("filter_macmode", "disabled")) {
		memset(&filter, 0, sizeof(filter));
		filter.dir = NETCONF_FORWARD;
		strcpy(filter.match.in.name, nvram_safe_get("lan_ifname"));
		if (nvram_match("filter_macmode", "allow")) {
			/* Allow new connections from the LAN side only from the specified addresses */
			filter.target = log_accept;
			filter.match.state = NETCONF_NEW;
		} else {
			/* Drop connections from the specified addresses */
			filter.target = log_drop;
		}
		foreach(var, nvram_safe_get("filter_mac"), next) {
			if (ether_atoe(var, filter.match.mac.octet))
				add_filter(&filter);
		}
		for (i = 0; i <= 20; i++) {
			snprintf(name, sizeof(name), "filter_mac%d", i);
			if (!*(value = nvram_safe_get(name)))
				continue;
			if ((strstr(value, "allow") || strstr(value, "deny")) &&
			    !strstr(value, nvram_safe_get("filter_macmode")))
				continue;
			if (ether_atoe(value, filter.match.mac.octet))
				add_filter(&filter);
		}
	}

	/*
	 * Inbound accepts
	 */

	/* Allow new connections from the LAN side */
	if (!nvram_match("filter_macmode", "allow")) {
		memset(&filter, 0, sizeof(filter));
		filter.match.state = NETCONF_NEW;
		strcpy(filter.match.in.name, nvram_safe_get("lan_ifname"));
		filter.target = log_accept;
		add_filter(&filter);
	}

	/* Allow new connections from the WAN side */
	if (nvram_match("fw_disable", "1")) {
		memset(&filter, 0, sizeof(filter));
		filter.match.state = NETCONF_NEW;
		strcpy(filter.match.in.name, nvram_safe_get("wan_ifname"));
		filter.target = log_accept;
		add_filter(&filter);
	}

	/* Allow established and related outbound connections back in */
	memset(&filter, 0, sizeof(filter));
	filter.match.state = NETCONF_ESTABLISHED | NETCONF_RELATED;
	filter.target = log_accept;
	add_filter(&filter);

	/*
	 * NAT
	 */

	/* Enable IP masquerading */
	if ((fp = fopen("/proc/sys/net/ipv4/ip_forward", "r+"))) {
		fputc('1', fp);
		fclose(fp);
	} else
		perror("/proc/sys/net/ipv4/ip_forward");

	/* ChungLiang added, for switch between router and gateway mode */
	if ( (c=nvram_get("WorkingModel")) != NULL)
		WorkingModel = c;
	if (*WorkingModel == '0') {	
		memset(&nat, 0, sizeof(nat));
//		inet_aton(nvram_safe_get("lan_ipaddr"), &nat.match.src.ipaddr);
//		inet_aton(nvram_safe_get("lan_netmask"), &nat.match.src.netmask);
		inet_aton("0.0.0.0", &nat.match.src.ipaddr);
		inet_aton("0.0.0.0", &nat.match.src.netmask);
		nat.match.src.ipaddr.s_addr &= nat.match.src.netmask.s_addr;
		strcpy(nat.match.out.name, nvram_safe_get("wan_ifname"));
		nat.target = NETCONF_MASQ;
		netconf_add_nat(&nat);
	}
	/* ChungLiang added end */

	/* Enable remote management */
/*	if ( nvram_invmatch ( "http_wanport" , "" ) ) {
		snprintf ( name , sizeof ( name ) , "%d-%d>%s:%d-%d" ,
			 atoi ( nvram_safe_get ( "http_wanport" ) ) ,
			 atoi ( nvram_safe_get ( "http_wanport" ) ) ,
			 nvram_safe_get ( "lan_ipaddr" ) ,
			 atoi ( nvram_safe_get ( "http_lanport" ) ) ,
			 atoi ( nvram_safe_get ( "http_lanport" ) ) );
		parse_port_forward ( IPPROTO_TCP , NETCONF_IN , log_accept , name );
	}
*/
	/* Forward TCP and UDP connections */
	parse_port_forward(IPPROTO_TCP, NETCONF_FORWARD, log_accept, nvram_safe_get("forward_tcp"));
	parse_port_forward(IPPROTO_UDP, NETCONF_FORWARD, log_accept, nvram_safe_get("forward_udp"));
	for (i = 0; i <= 20; i++) {
		snprintf(name, sizeof(name), "forward_port%d", i);
		parse_port_forward(0, NETCONF_FORWARD, log_accept, nvram_safe_get(name));
	}
/*	if (nvram_invmatch("dmz_ipaddr", "")) {
		snprintf(name, sizeof(name), "0-65535>%s:0-65535",
			 nvram_safe_get("dmz_ipaddr"));
		parse_port_forward(IPPROTO_TCP, NETCONF_FORWARD, log_accept, name);
		parse_port_forward(IPPROTO_UDP, NETCONF_FORWARD, log_accept, name);
	}	*/

	/* Clamp TCP MSS to PMTU of WAN interface */
	if (strcmp(nvram_safe_get("wan_proto"), "pppoe") == 0)
		netconf_clamp_mss_to_pmtu();

	/*
	 * Inbound defaults
	 */

	/* Log refused connections */
	memset(&filter, 0, sizeof(filter));
	filter.target = log_drop;
	add_filter(&filter);

	dprintf("done\n");
	return 0;
}

int
stop_firewall(void)
{
	dprintf("done\n");
	return 0;
}
