/*
 * Copyright 2001-2005, Broadcom Corporation
 * All Rights Reserved.
 * 
 * THIS SOFTWARE IS OFFERED "AS IS", AND BROADCOM GRANTS NO WARRANTIES OF ANY
 * KIND, EXPRESS OR IMPLIED, BY STATUTE, COMMUNICATION OR OTHERWISE. BROADCOM
 * SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A SPECIFIC PURPOSE OR NONINFRINGEMENT CONCERNING THIS SOFTWARE.
 *
 * $Id$
 *
*/
 
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <syslog.h>
#include <signal.h>
#include <string.h>
#include <sys/klog.h>
#include <sys/types.h>
#include <sys/mount.h>
#include <sys/reboot.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <sys/wait.h>
#include <dirent.h>

#include <epivers.h>
#include <bcmnvram.h>
#include <mtd.h>
#include <shutils.h>
#include <rc.h>
#include <netconf.h>
#include <nvparse.h>
#include <bcmdevs.h>

#include <sys/socket.h>
#include <sys/sysinfo.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <net/if_arp.h>

#include <wlioctl.h>
#include <wlutils.h>
#include <sys/ioctl.h>

#include "./gemtek_added/ExecutingTable.h"		/* Added by PeterChen */
#include "./gemtek_added/ConvParas.h"		/* Added by PeterChen */
#include "./gemtek_added/gemtek.h"
#include "./gemtek_added/Peter_Added_001.h"
#include "./gemtek_added/Dump_DHCP_Lease_Table.h"
	
char	wan_done_flag = 0;	/* Added by PeterChen, Gemtek 2003-0702 PM 08:37 */
void main_loop_stop_1(void);
void main_loop_stop_2(void);
void main_loop_start_1(void);
void main_loop_start_2(void);
static void restore_defaults(void);
static void sysinit ( void );
static void rc_signal(int sig);


/* Walter Add 2004-08-20 */
 #include <wlutils.h>
 #define  WalterDBG(x)  x
/* Walter Add 2004-08-20 end*/

extern struct nvram_tuple	router_defaults[];

static void
restore_defaults(void)
{
#if 0
	struct nvram_tuple generic[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "eth0 eth2 eth3 eth4", 0 },
		{ "wan_ifname", "eth1", 0 },
		{ "wan_ifnames", "eth1", 0 },
		{ 0, 0, 0 }
	};
	struct nvram_tuple vlan[] = {
		{ "lan_ifname", "br0", 0 },
		{ "lan_ifnames", "vlan0 eth1 eth2 eth3", 0 },
		{ "wan_ifname", "vlan1", 0 },
		{ "wan_ifnames", "vlan1", 0 },
		{ 0, 0, 0 }
	};
	struct nvram_tuple *linux_overrides;
	struct nvram_tuple *t, *u;
	int restore_defaults, i;
	uint boardflags;

	/* Restore defaults if told to or OS has changed */
	restore_defaults = !nvram_match("restore_defaults", "0") || nvram_invmatch("os_name", "linux");
	if (restore_defaults)
		cprintf("Restoring defaults...");

	/* Delete dynamically generated variables */
	if (restore_defaults) {
		char tmp[100], prefix[] = "wlXXXXXXXXXX_";
		for (i = 0; i < MAX_NVPARSE; i++) {
			del_filter_client(i);
			del_forward_port(i);
			del_autofw_port(i);
			snprintf(prefix, sizeof(prefix), "wl%d_", i);
			for (t = router_defaults; t->name; t ++) {
				if (!strncmp(t->name, "wl_", 3))
					nvram_unset(strcat_r(prefix, &t->name[3], tmp));
			}
			snprintf(prefix, sizeof(prefix), "wan%d_", i);
			for (t = router_defaults; t->name; t ++) {
				if (!strncmp(t->name, "wan_", 4))
					nvram_unset(strcat_r(prefix, &t->name[4], tmp));
			}
		}
	}

	/* Choose default lan/wan i/f list. */
	boardflags = strtoul(nvram_safe_get("boardflags"), NULL, 0);
	if (boardflags & BFL_ENETVLAN)
		linux_overrides = vlan;
	else
		linux_overrides = generic;

	/* Restore defaults */
	for (t = router_defaults; t->name; t++) {
		if (restore_defaults || !nvram_get(t->name)) {
			for (u = linux_overrides; u && u->name; u++) {
				if (!strcmp(t->name, u->name)) {
					nvram_set(u->name, u->value);
					break;
				}
			}
			if (!u || !u->name)
				nvram_set(t->name, t->value);
		}
	}

	/* Always set OS defaults */
	nvram_set("os_name", "linux");
	nvram_set("os_version", EPI_ROUTER_VERSION_STR);

	/* Commit values */
	if (restore_defaults) {
		nvram_commit();
		cprintf("done\n");
	}
#endif	
}

static void
set_wan0_vars(void)
{
	int unit;
	char tmp[100], prefix[] = "wanXXXXXXXXXX_";
	
	/* check if there are any connections configured */
	for (unit = 0; unit < MAX_NVPARSE; unit ++) {
		snprintf(prefix, sizeof(prefix), "wan%d_", unit);
		if (nvram_get(strcat_r(prefix, "unit", tmp)))
			break;
	}
	/* automatically configure wan0_ if no connections found */
	if (unit >= MAX_NVPARSE) {
		struct nvram_tuple *t;
		char *v;

		/* Write through to wan0_ variable set */
		snprintf(prefix, sizeof(prefix), "wan%d_", 0);
		for (t = router_defaults; t->name; t ++) {
			if (!strncmp(t->name, "wan_", 4)) {
				if (nvram_get(strcat_r(prefix, &t->name[4], tmp)))
					continue;
				v = nvram_get(t->name);
				nvram_set(tmp, v ? v : t->value);
			}
		}
		nvram_set(strcat_r(prefix, "unit", tmp), "0");
		nvram_set(strcat_r(prefix, "desc", tmp), "Default Connection");
		nvram_set(strcat_r(prefix, "primary", tmp), "1");
	}
}

static int		noconsole = 0;

static void	
sysinit ( void )
{
	char		buf [ PATH_MAX ];
	struct utsname	name;
	struct stat	tmp_stat;
	time_t		tm = 0;

	/* /proc */
	mount ( "proc" , "/proc" , "proc" , MS_MGC_VAL , NULL );

	/* /tmp */
	mount ( "ramfs" , "/tmp" , "ramfs" , MS_MGC_VAL, NULL );

	/* /var */
	mkdir ( "/tmp/var" , 0777 );
	mkdir ( "/var/lock" , 0777 );
	mkdir("/var/log", 0777);
	mkdir("/var/run", 0777);
	mkdir("/var/tmp", 0777);
	

	/* Setup console */
	if ( console_init () )
		noconsole = 1;
	klogctl ( 8 , NULL , atoi ( nvram_safe_get ( "console_loglevel" ) ) );

	/* Get CPU info */
	check_brcm_cpu_type ();

	convert_wireless_params ();

	/* Modules */
	uname ( &name );
	snprintf ( buf , sizeof ( buf ) , "/lib/modules/%s" , name.release );
	if (stat("/proc/modules", &tmp_stat) == 0 &&
	    stat(buf, &tmp_stat) == 0) {
		char module[80], *modules, *next;
		//modules = nvram_get("kernel_mods") ? : "et il wl";
		/*Yao Chung Added*/
		char *modulestr;
		if ((atoi(nvram_safe_get("WlEnable")) == 0 ) && (atoi(nvram_safe_get("Wl5GEnable")) == 0 ))
			modulestr = "kmemory et";
		else if (atoi(nvram_safe_get("Wl5GEnable")) == 0 )
			modulestr = "kmemory et wl";
		else if	(atoi(nvram_safe_get("WlEnable")) == 0 )
		{
			//Remove by Terence, 20030507
			modulestr = "kmemory et";
			//modulestr = "kmemory et tulip tulip_1";
		}
		else
		{
			//Remove by Terence, 20030507
			modulestr = "kmemory et wl";
			//modulestr = "kmemory et wl tulip tulip_1";
		}
		modules = modulestr;
		Peter_Added_001_H_Extra_SRCs_001

		/*Yao Chung Added*/	
		/*modules = nvram_get("kernel_mods") ? : "kmemory et il wl tulip tulip_1";
		*/
		foreach(module, modules, next)
			eval("insmod", module);
//		Peter_Added_001_H_Extra_SRCs_003
	}

	/* Set a sane date */
	stime ( &tm );

/*Yuan : added for intersil isl39xx (11a or 11g) module */
#if	0
{
	char	buf [ 50 ] , *ssid , *channel;
	ssid = nvram_get ( "_SSID_" );
	channel = nvram_get ( "_Channel_" );

	if ( ssid != NULL && channel != NULL ) {
		sprintf(buf , "/sbin/insmod isl38xx init_ssid=\"%s\" init_channel=%s", ssid, channel) ;
	} else if (ssid != NULL && channel == NULL) {
		sprintf(buf , "/sbin/insmod isl38xx init_ssid=\"%s\"", ssid) ;
	} else if (ssid == NULL && channel != NULL) {
		sprintf(buf , "/sbin/insmod isl38xx init_channel=%s", channel) ;
	} else { // ssid == NULL && channel == NULL 
		sprintf(buf , "/sbin/insmod isl38xx") ;
	}

	system ( buf ) ;	/* rc_system(buf); */
}
#endif

/*End of Yuan*/

/* oscar: Added for dproxy resolv.conf. 2002.10.2 */
#if	0
{
	FILE *fp;
	char buf[50], *nvbuf;
	int i = 0;
	fp = fopen("/tmp/resolv.conf", "w");
	if(fp != NULL) {     
	  nvbuf = nvram_get("wan_dns");
	  if(nvbuf != NULL) { // Have to check if wan_dns is existed, otherwise kernel will crash. 2002.10.09

	    while(buf[i] = *nvbuf++) {
	    	if(buf[i] == ' ') {
		  buf[i] = '\0';
		  fprintf(fp, "nameserver %s\n", buf);
		  i = 0;
		}
		else
		     	i++;
	    }

	    if(strlen(buf))    /* Not empty */
		fprintf(fp, "nameserver %s\n", buf);

	    fflush(fp);
	  }		
		fclose(fp);
	}
}
#endif

/* end of oscar*/

#if	0 // marked by Jimmy Ma 2003.6.18
/* Andrew added to PPTP 2003.01.02 */
{
	FILE	*sd;
	char 	usr[20], *usrname, *pwd, *ip, *idletm;
	char	*strrcv;
	int		len;
	int		demand;

	bzero(usr, 20);
	strrcv = nvram_safe_get("pptp_demand");

	//printf("Get parameter pptp_demand from flash \n");
	if (strrcv != NULL){
		demand	= atoi(strrcv);
		idletm  = nvram_get("pptp_idletime");
		if (idletm == NULL)		goto end;
		usrname = nvram_get("pptp_username");
		if (usrname == NULL)	goto end;
		pwd = nvram_get("pptp_passwd");
		if (pwd == NULL)		goto end;
		usr[0] = '"';
		strcpy(&usr[1], usrname);
		strncat(usr , &usr[0],1);
	}
	else	goto end;

	sd = fopen("/tmp/options", "w");
	if (sd != NULL)
	{
		fprintf(sd,"name %s\n",usr);
		fprintf(sd,"noauth \n");
		fprintf(sd,"noipdefault \n");
		fprintf(sd,"defaultroute \n");
		fprintf(sd,"debug \n");
		fprintf(sd,"lock \n");
			if (demand ==1)
				fprintf(sd,"idle %s \n",idletm);
		fclose(sd);
	}
	else
		fclose(sd);

	ip = nvram_get("pptp_gateway");
	sd = fopen("/tmp/pap-secrets", "w");
	if (sd != NULL){
		// printf("Successfully open /tmp/pap-secrets \n");
		fprintf(sd, "%s	   %s	%s\n",usrname, ip, pwd);
		fclose(sd);
	}
	else
		fclose(sd);
end:
	// printf("End of RC open file \n");
}
/* end of Andrew */
#endif

	/*Add by Yao Chung
	symlink("/sbin/rc", "/tmp/start_wan");
	symlink("/sbin/rc", "/tmp/stop_wan");*/
	symlink("/sbin/rc", "/tmp/GetPPTPAddr");
	/*Add by Yao Chung*/
	dprintf("done\n");
      

}

/* States */
enum {
	RESTART,
	STOP,
	START,
	TIMER,
	IDLE,
};
static int state = START;
static int signalled = -1;

/* Signal handling */
static void
rc_signal(int sig)
{
	if (state == IDLE) {	
		if (sig == SIGHUP) {
			dprintf("signalling RESTART\n");
			signalled = RESTART;
		}
		else if (sig == SIGUSR2) {
			dprintf("signalling START\n");
			signalled = START;
		}
		else if (sig == SIGINT) {
			dprintf("signalling STOP\n");
			signalled = STOP;
		}
		else if (sig == SIGALRM) {
			dprintf("signalling TIMER\n");
			signalled = TIMER;
		}
	}
}

/* Timer procedure */
int
do_timer(void)
{
	int interval = atoi(nvram_safe_get("timer_interval"));
	time_t now;
	struct tm gm, local;
	struct timezone tz;

	dprintf("%d\n", interval);

	if (interval == 0)
		return 0;

	/* Report stats */
	if (nvram_invmatch("stats_server", "")) {
		char *stats_argv[] = { "stats", nvram_get("stats_server"), NULL };
		_eval(stats_argv, NULL, 5, NULL);
	}

	/* Sync time */
	//start_ntpc();

	/* Update kernel timezone */
	setenv("TZ", nvram_safe_get("time_zone"), 1);
	time(&now);
	gmtime_r(&now, &gm);
	localtime_r(&now, &local);
	tz.tz_minuteswest = (mktime(&gm) - mktime(&local)) / 60;
	settimeofday(NULL, &tz);

	alarm(interval);
	return 0;
}

/* Timer procedure - Kings */
//int
//do_timer(void)
//{
//	int interval = atoi(nvram_safe_get("timer_interval"));
//
//	dprintf("%d\n", interval);
//
//	if (interval == 0)
//		return 0;
//
//	/* Report stats */
//	if (nvram_invmatch("stats_server", "")) {
//		char *stats_argv[] = { "stats", nvram_get("stats_server"), NULL };
//		_eval(stats_argv, NULL, 5, NULL);
//	}
//
//	/* Sync time */
//	if (nvram_invmatch("ntp_server", "")) {
//		char *ntpclient_argv[] = { "ntpclient", "-h", nvram_get("ntp_server"), "-l", "-s", NULL };
//		_eval(ntpclient_argv, NULL, 5, NULL);
//	}
//
//	alarm(interval);
//	return 0;
//}

/* Main loop */
static void	
main_loop ( void )
{
	sigset_t	sigset;
	pid_t	shell_pid = 0;
	uint boardflags;
	
	/* Basic initialization */
	sysinit ();

	// convert_wireless_params ();		/* Nvram convert for RC48.7 */
	// check_nvram_parameters ();
	// convert_wireless_params ();

	/* Setup signal handlers */
	signal_init ();
	signal ( SIGHUP , rc_signal );
	signal ( SIGUSR2 , rc_signal );
	signal ( SIGINT , rc_signal );
	signal ( SIGALRM , rc_signal );
	sigemptyset ( &sigset );

	/* Give user a chance to run a shell before bringing up the rest of the system */
	//if (!noconsole)
	//	run_shell(1, 0);
	
	/* Convert deprecated variables */
	//convert_deprecated();	
	
	/* Restore defaults if necessary */
	//restore_defaults();	
	
	/* Marked by YaoChung*/
	gemtek_restore_defaults();
	// start_upgrade_pmon();
	/*Marked by YaoChung*/

	/* Setup wan0 variables if necessary */
	set_wan0_vars();
	/* Add loopback */
	config_loopback();
	
	/* Add vlan */
	boardflags = strtoul(nvram_safe_get("boardflags"), NULL, 0);
	if (boardflags & BFL_ENETVLAN)
		config_vlan();
 	/* Loop forever */
	for ( ; ; ) {
		switch ( state ) {
		case	RESTART:
			dprintf ( "RESTART\n" );
			/* Fall through */
		case	STOP:
			dprintf ( "STOP\n" );
			//stop_services();
			main_loop_stop_1();
			stop_wan ();
			main_loop_stop_2();
			sleep ( 2 );
			//stop_lan();
			if ( state == STOP ) {
				state = IDLE;
				break;
			}
			/* Fall through */
		case START:
			//dprintf("START\n");
			main_loop_start_1();
			start_lan ();
			
		      /* Walter add 2004-08-23 for WDS*/		        
		       // WalterDBG(fprintf ( stderr , "rc.c L558: start wlconfig---\n");)
		        /*  
		           char name[80], *next;		        
		         
                           foreach ( name , nvram_safe_get ( "lan_ifnames" ) , next ) {		        		       		       
                           int unit;
		           wl_ioctl(name, WLC_GET_INSTANCE, &unit, sizeof(unit));
		           WalterDBG(fprintf ( stderr , "wl_ioctl: %s, result: %d\n",name,unit);) 
		           memset(name,0,sizeof(name));
		           if(unit !=0)
		           {
		             WalterDBG(fprintf ( stderr , "Start delay...\n");) 
		             usleep(100000);
		           }
                        }
                      */  
                       //add code to check to prevent page fault here
		        
		       // eval("wlconfig",nvram_safe_get("wlif"),"up"); 
		       // WalterDBG(fprintf ( stderr , "rc.c L561: end wlconfig eth1\n");)
		        /* Walter add for WDS end*/
		        
			   //start_services();
			  
			/* start_wan and start_nas move to main_loop_start_2() */
			//start_wan();
			//start_nas("wan");			
			main_loop_start_2();
		case TIMER:
			dprintf("TIMER\n");
			do_timer();
			/* Fall through */
		case IDLE:
			dprintf("IDLE\n");
			state = IDLE;
			/* Wait for user input or state change */
			while (signalled == -1) {
				if (!noconsole && (!shell_pid || kill(shell_pid, 0) != 0))
					shell_pid = run_shell(0, 1);
				else
					sigsuspend(&sigset);
			}
			state = signalled;
			signalled = -1;
			break;
		default:
			dprintf("UNKNOWN\n");
			return;
		}
	}

	/* Give user a chance to run a shell before bringing up the rest of the system */
	if (!noconsole)
		run_shell (1 , 0);
}

int	
main ( int argc , char **argv )
{
	char	*base = strrchr ( argv[0] , '/' );

	base = base ? base + 1 : argv[0];
//	nvram_set ( "wan_lease" , "*******" );

	/* init */
	if ( strstr ( base , "init" ) ) {
		main_loop ();
		//return 0;
	} 
	
	/* Set TZ for all rc programs */
	//setenv("TZ", nvram_safe_get("time_zone"), 1);
	
	/* ppp */
	if ( strstr ( base , "ip-up" ) )	
		return	ipup_main ( argc , argv );
	else if ( strstr ( base , "ip-down" ) )
		return	ipdown_main ( argc , argv );
		
	/* udhcpc [ deconfig bound renew ] */
	else if ( strstr ( base , "udhcpc" ) )	
		return	udhcpc_main ( argc , argv );
		
	/* stats [ url ] */		
	else if ( strstr ( base , "stats" ) )	
		return	http_stats ( argv[1] ? : nvram_safe_get ( "stats_server" ) );

#ifdef MTD
	/* erase [device] */
	else if ( strstr ( base , "erase" ) ) {	
		if ( argv[1] )
			return	mtd_erase ( argv[1] );
		else {
			fprintf ( stderr , "usage: erase [device]\n" );
			return	EINVAL;
		}
	}
	
	/* write [path] [device] */
	else if ( strstr ( base , "write" ) ) {	
		if ( argc >= 3 )
			return	mtd_write ( argv[1] , argv[2] );
		else {
			fprintf ( stderr , "usage: write [path] [device]\n" );
			return	EINVAL;
		}
	}
#endif
	/* hotplug [event] */
	else if ( strstr ( base , "hotplug" ) ) {	
		if ( argc >= 2 ) {
			if ( !strcmp ( argv[1] , "net" ) )
				return	hotplug_net ();
		} else {
			fprintf ( stderr , "usage: hotplug [event]\n" );
			return	EINVAL;
		}
	} 
	
	/* rc [stop|start|restart ] */
	else if ( strstr ( base , "rc" ) ) {	
		if ( argv[1] ) {
			if ( strncmp ( argv[1] , "start" , 5 ) == 0 )
				return	kill ( 1 , SIGUSR2 );
			else if (strncmp(argv[1], "stop", 4) == 0)
				return	kill ( 1 , SIGINT );
			else if (strncmp(argv[1], "restart", 7) == 0)
				return	kill ( 1 , SIGHUP );
		} else {
			fprintf ( stderr , "usage: rc [start|stop|restart]\n" );
			return	EINVAL;
		}
	} 
	else if ( strstr ( base , "start_wan" ) )	/* start & stop wan connection */
		start_wan();
	else if ( strstr ( base , "stop_wan" ) )
		stop_wan();
	else if ( strstr ( base , "GetPPTPAddr" ) )/*get pptp address*/
		GetPPTPAddr();
	else if ( strstr ( base , "assoclistd" ) ) {		/* Modify by Terence from Dainel */
		start_wl_assoclist ();
	}
	else if ( strstr ( base , "sitesurvey" ) ) {
		wlconfig ( nvram_safe_get("wlif") );
	}		

	return EINVAL;
}

void rc_system ( const char *string )
{
	char	i = 10;

	if ( string == NULL ) { return; }

	while ( i > 0 ) {
		i--;
		if ( system ( string ) != -1 ) {
			if ( i < 10 ) { /* printf ( "Re-executing utility '%s' ... Success\n" , string ); */ }
			return;
		} else {
			/* printf ( "Executing utility '%s' ... Failure\n" , string ); */
			sleep ( 10 - i );	/* Added by PeterChen , Gemtek , 2003-0113-16:51 */
		}
	}

	printf ( "system(%s) fails to execute with errno=%d\n" , string , errno );
}

void main_loop_stop_1(void)
{
	netconf_reset_fw ();
	stop_firewall ();
	stop_nas();
}

void main_loop_stop_2(void)
{
		OpenFile ();	/* Added by PeterChen */
		WriteFile ( "/usr/bin/killall udhcpc" );
		WriteFile ( "/usr/bin/killall pppoecd" );
		//WriteFile ( "/usr/bin/killall easyconf" );
		WriteFile ( "/usr/bin/killall traceroute" );
		WriteFile ( "/usr/bin/killall exeNtp" );
		WriteFile ( "/usr/bin/killall Getntp" );
		WriteFile ( "/usr/bin/killall tproxy" );
		WriteFile ( "/usr/bin/killall ResetToDefault" );
		WriteFile ( "/usr/bin/killall ResetToDefault_4702" );
		WriteFile ( "/usr/bin/killall ResetToDefault_" );
//			WriteFile ( "/sbin/iptables -t filter -F" );
//			WriteFile ( "/sbin/iptables -t nat -F" );
//			WriteFile ( "/sbin/iptables -t mangle -F" );
		netconf_reset_fw ();
		WriteFile ( "/usr/bin/killall pptp" );
		WriteFile ( "/usr/bin/killall updatedd" );
		WriteFile ( "/usr/bin/killall conn_log" );
		WriteFile ( "/usr/bin/killall reaim" );
		WriteFile ( "/usr/bin/killall filter" );
		WriteFile ( "/usr/bin/killall portfw" );
		WriteFile ( "/usr/bin/killall assoclistd" );
		WriteFile ( "/usr/bin/killall udhcpd" );
		WriteFile ( "/usr/bin/killall zebra" );
		WriteFile ( "/usr/bin/killall ripd" );
		WriteFile ( "/usr/bin/killall dnsmasq" );
		WriteFile ( "/usr/bin/killall upnp" );
		WriteFile ( "/usr/bin/killall updatedd" );
		WriteFile ( "/usr/bin/killall udhcpd" );
		//WriteFile ( "/usr/bin/killall igmp-proxy" );
		WriteFile ( "/usr/bin/killall zebrip" ); //Walter add 2004-07-21		

		CloseFile ();	/* Added by PeterChen */
		Executing ();	/* Added by PeterChen */

//			stop_upnp ();
//			stop_dhcpd ();
//			stop_httpd ();	
}

void main_loop_start_1(void)
{
		Dump_DHCP_Lease_Table_H_Extra_SRCs_001
		OpenFile ();	/* Added by PeterChen */

		//nvram_set ( "WAN_FROM_GPIO" , "0" );
		nvram_set ( "pptp_ipaddr" , "0.0.0.0" );

		nvram_set ( "pppoe_ipaddr" , "0.0.0.0" );
		nvram_set ( "pppoe_netmask" , "0.0.0.0" );
		nvram_set ( "pppoe_gateway" , "0.0.0.0" );
		nvram_set ( "pppoe_dnsaddr1" , "0.0.0.0" );
		nvram_set ( "pppoe_dnsaddr2" , "0.0.0.0" );

		/* YaoChung Add for dhcp/pppoe*/
		{
			char	*wan_proto = NULL;
			int	StringLen = 0;

			wan_proto = nvram_safe_get ( "wan_proto" );
			if ( wan_proto != NULL ) {
				StringLen = strlen ( wan_proto );
				if ( StringLen != 0 ) {
					if ( strcmp ( wan_proto , "dhcp" ) == 0 ) {
						nvram_set ( "wan_lease" , "*******" );
						goto	BaBa;
					}

					if ( strcmp ( wan_proto , "pppoe" ) == 0 ) {
BaBa:							nvram_set ( "wan_ipaddr" , "0.0.0.0" );
						nvram_set ( "wan_netmask" , "0.0.0.0" );
						nvram_set ( "wan_gateway" , "0.0.0.0" );
						nvram_set ( "wan_dns" , "0.0.0.0 0.0.0.0 0.0.0.0" );						
					}
				}
			}
		}
}

void main_loop_start_2(void)
{
	char	*wan_gateway = NULL;		// andrew add 20030529
	char	string[40];			// andrew add 20030529

	memset ( string , '\0' , 40 );
	
			/* Start firewall */
			start_firewall ();

			{	/* Added by PeterChen, Gemtek 2003-0704 PM 01:48 */
				/* For speeding up the system rebooting */
//				chdir ( "/sbin" );
//				eval ( "/sbin/udhcpdset" );
//				chdir ( "/" );

				char		*WorkingModel = nvram_get ( "WorkingModel" );

				if ( WorkingModel != NULL ) {
					if ( *WorkingModel == '1' ) {			/* Router mode */
						rc_system ( "/sbin/udhcpdset;/sbin/iptables -t nat -I PREROUTING 1 -i br0 -p udp --dport 67 -j ACCEPT;/sbin/iptables -t filter -I FORWARD 1 -j ACCEPT;/sbin/iptables -t filter -I INPUT 1 -i br0 -p udp --dport 67 -j ACCEPT;" );
					} else {							/* Gateway mode */
						rc_system ( "/sbin/udhcpdset;/sbin/iptables -t nat -I PREROUTING 1 -i br0 -p udp --dport 67 -j ACCEPT;/sbin/iptables -t filter -I INPUT 1 -i br0 -p udp --dport 67 -j ACCEPT;" );
					}
				} else {
					rc_system ( "/sbin/udhcpdset;/sbin/iptables -t nat -I PREROUTING 1 -i br0 -p udp --dport 67 -j ACCEPT;/sbin/iptables -t filter -I INPUT 1 -i br0 -p udp --dport 67 -j ACCEPT;" );
				}

//				rc_system ( "/sbin/udhcpdset;/sbin/iptables -t nat -I PREROUTING 1 -i br0 -p udp --dport 67 -j ACCEPT;/sbin/iptables -t filter -I FORWARD 1 -i br0 -p udp --dport 67 -j ACCEPT;/sbin/iptables -t filter -I INPUT 1 -i br0 -p udp --dport 67 -j ACCEPT;" );

/*				pid_t	pidXXX;
				char	*udhcpdset_argv[] = { "/sbin/udhcpdset" , NULL };
				_eval ( udhcpdset_argv , NULL , 0 , &pidXXX );	*/

				// rc_system ( "/sbin/udhcpdset &" );	/* Added by PeterChen, Gemtek 2003-0702 PM 01:39 */
			}
			// rc_system ( "/sbin/iptables -t filter -I FORWARD 1 -j ACCEPT;/sbin/iptables -t nat -I PREROUTING 1 -j ACCEPT;/sbin/iptables -t filter -I INPUT 1 -j ACCEPT;" );

			// sleep ( 3 );

			start_httpd ();

//
			// sleep ( 10 );



			// start_dhcpd ( 60 );

			/*printf("ConnectionOption value = %s\n",nvram_safe_get("ConnectionOption"));*/
{
			char	*Gemtek = nvram_safe_get ( "ConnectionOption" );
			int	GemtekStringLen = 0;

			if ( Gemtek != NULL ) {
				GemtekStringLen = strlen ( Gemtek );
				if ( GemtekStringLen != 0 ) {
					if ( atoi ( Gemtek ) != 1 ) {
						start_wan ();
					} else {
						start_wan_done ();

						/* Modified by PeterChen, Gemtek 2003-0702 PM 02:39 */
						// WriteFile ( "/sbin/filter &" );

						Gemtek = nvram_safe_get ( "wan_proto" );
						if ( Gemtek != NULL ) {
							if ( strcasecmp ( Gemtek , "dhcp" ) == 0 ) {
								rc_system ( "/usr/bin/killall udhcpc" );
								// WriteFile ( "/usr/bin/killall udhcpc &" );
							}
    
						}
					}
				}
			}
}

//			sleep ( 10 );

			nvram_set ( "ConnectionOption" , "0" );

			/* YaoChung 2002.12.13 for snmp log pptp DDNS wirelessMAC*/
			{	
				/*after filter nvram_set is no use*/
				GetWirelessMAC ();

				gemtek_restore_defaults_with_reboot();

				DDNSSet ();

				/* YaoChung 2002.12.13 for snmp log pptp DDNS wireless*/
				/* oscar, 2002.10.02 */

				// rc_system("/sbin/dproxy");
				// WriteFile ( "/sbin/dproxy -d &" );
				// WriteFile ( "dnsmasq -a 3.3.3.1 -i br0 -l /tmp/udhcpd.leases -q -r /tmp/resolv.conf -x /tmp/dnsmasq.pid" );

				/*rc_system("/sbin/parentCnt&");*/

				WriteFile ( "/sbin/reaim &" );
				// rc_system("/sbin/reaim&");			

				/* ChungLiang start, 2002.10.25 */
				// rc_system("echo \"1\" > /proc/sys/net/ipv4/icmp_echo_ignore_broadcasts");
				WriteFile ( "echo \"1\" > /proc/sys/net/ipv4/icmp_echo_ignore_broadcasts" );

				WriteFile ( "/sbin/zebrip &" );

				//Mark to prevent from DHCP crash Terence 20030513
				//start_firewall();
				// rc_system("/sbin/easyconf&");
				//WriteFile ( "/sbin/easyconf &" );

				/*Added by Peter*/	
				// rc_system ( "/sbin/udhcpdset" );
				// WriteFile ( "/sbin/udhcpdset &" );

				/* Add by Terence */
				// rc_system( "/sbin/assoclistd &" );
				WriteFile ( "/sbin/assoclistd &" );
				/* rc_system ( "sleep 3" ); */
			}

//			if ( wan_done_flag == 1 ) {
				WriteFile ( "/sbin/exeNtp &" );
				// WriteFile ( "/sbin/filter &" );
				rc_system ( "/sbin/filter &" );
			#if 0								
			//Walter add, 2004-09-03 for allow PPPoE can carry on UPnP
                          if(!strcmp("pppoe",nvram_safe_get("wan_proto")))
	                  {
 	                    if ( !strcmp ( nvram_safe_get ( "UPnPEnable" ) , "1" ) ) 
                              rc_system ( "/usr/sbin/upnp &" ); 			    
	                  } 
	                #endif  
//			}

			Peter_Added_001_H_Extra_SRCs_002 //Walter: check board and load 4712 or 4702 ResetToDefault
			// rc_system ( "/sbin/ResetToDefault&" );
			// WriteFile ( "/sbin/ResetToDefault &" );

			// WriteFile ( "/sbin/dnsmasq -i br0 -l /tmp/udhcpd.leases -q -r /tmp/resolv.conf -x /tmp/dnsmasq.pid &" );

			/* eval ( "udhcpd" , "/tmp/udhcpd.conf" ); */
			/*Added by Peter*/
#if	0
			char    *lan_ipaddr = NULL , *lan_netmask = NULL , buf [ 50 ];

 			lan_ipaddr = nvram_get ( "lan_ipaddr" );
 			if ( ( lan_ipaddr == NULL ) || ( strlen ( lan_ipaddr ) > strlen ( "xxx.xxx.xxx.xxx" ) ) ) {
  				lan_ipaddr = "192.168.10.1";
 			}

 			lan_netmask = nvram_get ( "lan_netmask" );
 			if ( ( lan_netmask == NULL ) || ( strlen ( lan_netmask ) > strlen ( "xxx.xxx.xxx.xxx" ) ) ) {
  			lan_netmask = "255.255.255.0";
 			}

			sprintf ( buf , "/sbin/iptables -t nat -I PREROUTING 1 -p tcp -s %s/%s -d %s --dport 80 -j ACCEPT &\n" , lan_ipaddr , lan_netmask , lan_ipaddr );
			rc_system ( buf );
			WriteFile ( buf );
#endif

			/* ChungLiang end, 2002.10.25 */
			/* Fall through */
			/* Modify by Terence for Motorola */
			if ( !strcmp ( nvram_safe_get ( "WAN_ENABLE" ) , "1" ) ) {
				//printf("rc : ifconfig eth1 up...Do nothing\n");
				LED_wan_if_enable ( 1 );
				//system("ifconfig eth1 up");
			} else {
				//printf("rc : ifconfig eth1 down\n");
				LED_wan_if_enable ( 0 );
				//system("ifconfig eth1 down");			
				ifconfig ( nvram_safe_get ( "wan_ifname" ) , 0 , NULL , NULL );
			}			

			/* Add by Terence for UPnP 20030516 */
//			if ( !strcmp ( nvram_safe_get ( "UPnPEnable" ) , "1" ) ) {
				// rc_system("/usr/sbin/upnp");
//				WriteFile ( "/usr/sbin/upnp &" );
//			}

			CloseFile ();	/* Added by PeterChen */
			Executing ();	/* Added by PeterChen */

			start_nas ("lan");	
			
}
